#!/usr/bin/env python3.6
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"command line module"

from __future__ import print_function

import argparse
import ujson as json
import logging
import os
import sys
import traceback
import yaml

from kapitan.utils import jsonnet_file, PrettyDumper, flatten_dict, searchvar, deep_get, check_version, save_version
from kapitan.targets import compile_targets
from kapitan.resources import search_imports, resource_callbacks, inventory_reclass
from kapitan.version import PROJECT_NAME, DESCRIPTION, VERSION
from kapitan.secrets import secret_gpg_backend, secret_gpg_write, secret_gpg_reveal_file
from kapitan.secrets import secret_gpg_reveal_dir, secret_gpg_reveal_raw, secret_gpg_update_recipients
from kapitan.secrets import search_target_token_paths, secret_gpg_raw_read_fingerprints
from kapitan.secrets import lookup_fingerprints
from kapitan.errors import KapitanError

logger = logging.getLogger(__name__)


def main():
    "main function for command line usage"
    parser = argparse.ArgumentParser(prog=PROJECT_NAME,
                                     description=DESCRIPTION)
    parser.add_argument('--version', action='version', version=VERSION)
    subparser = parser.add_subparsers(help="commands")

    eval_parser = subparser.add_parser('eval', help='evaluate jsonnet file')
    eval_parser.add_argument('jsonnet_file', type=str)
    eval_parser.add_argument('--output', type=str,
                             choices=('yaml', 'json'), default='yaml',
                             help='set output format, default is "yaml"')
    eval_parser.add_argument('--vars', type=str, default=[], nargs='*',
                             metavar='VAR',
                             help='set variables')
    eval_parser.add_argument('--search-path', '-J', type=str, default='.',
                             metavar='JPATH',
                             help='set search path, default is "."')

    compile_parser = subparser.add_parser('compile', help='compile targets')
    compile_parser.add_argument('--search-path', '-J', type=str, default='.',
                                metavar='JPATH',
                                help='set search path, default is "."')
    compile_parser.add_argument('--verbose', '-v', help='set verbose mode',
                                action='store_true', default=False)
    compile_parser.add_argument('--no-prune', help='do not prune jsonnet output',
                                action='store_true', default=False)
    compile_parser.add_argument('--quiet', help='set quiet mode, only critical output',
                                action='store_true', default=False)
    compile_parser.add_argument('--output-path', type=str, default='.',
                                metavar='PATH',
                                help='set output path, default is "."')
    compile_parser.add_argument('--targets', '-t', help='targets to compile, default is all',
                                type=str, nargs='+', default=[], metavar='TARGET')
    compile_parser.add_argument('--parallelism', '-p', type=int,
                                default=4, metavar='INT',
                                help='Number of concurrent compile processes, default is 4')
    compile_parser.add_argument('--indent', '-i', type=int,
                                default=2, metavar='INT',
                                help='Indentation spaces for YAML/JSON, default is 2')
    compile_parser.add_argument('--secrets-path', help='set secrets path, default is "./secrets"',
                                default='./secrets',)
    compile_parser.add_argument('--reveal',
                                help='reveal secrets (warning: this will write sensitive data)',
                                action='store_true', default=False)
    compile_parser.add_argument('--inventory-path', default='./inventory',
                                help='set inventory path, default is "./inventory"')
    compile_parser.add_argument('--ignore-version-check',
                                help='ignore the last kapitan version used to compile (from .kapitan)',
                                action='store_true', default=False)

    inventory_parser = subparser.add_parser('inventory', help='show inventory')
    inventory_parser.add_argument('--target-name', '-t', default='',
                                  help='set target name, default is all targets')
    inventory_parser.add_argument('--inventory-path', default='./inventory',
                                  help='set inventory path, default is "./inventory"')
    inventory_parser.add_argument('--flat', '-F', help='flatten nested inventory variables',
                                  action='store_true', default=False)
    inventory_parser.add_argument('--pattern', '-p', default='',
                                  help='filter pattern (e.g. parameters.mysql.storage_class, or storage_class,' +
                                  ' or storage_*), default is ""')

    searchvar_parser = subparser.add_parser('searchvar',
                                            help='show all inventory files where var is declared')
    searchvar_parser.add_argument('searchvar', type=str, metavar='VARNAME',
                                  help='e.g. parameters.mysql.storage_class, or storage_class, or storage_*')
    searchvar_parser.add_argument('--inventory-path', default='./inventory',
                                  help='set inventory path, default is "./inventory"')

    secrets_parser = subparser.add_parser('secrets', help='manage secrets')
    secrets_parser.add_argument('--write', '-w', help='write secret token',
                                metavar='TOKENNAME',)
    secrets_parser.add_argument('--update', help='update recipients for secret token',
                                metavar='TOKENNAME',)
    secrets_parser.add_argument('--update-targets', action='store_true', default=False,
                                help='update target secrets')
    secrets_parser.add_argument('--validate-targets', action='store_true', default=False,
                                help='validate target secrets')
    secrets_parser.add_argument('--base64', '-b64', help='base64 encode file content',
                                action='store_true', default=False)
    secrets_parser.add_argument('--reveal', '-r', help='reveal secrets',
                                action='store_true', default=False)
    secrets_parser.add_argument('--file', '-f', help='read file or directory, set "-" for stdin',
                                metavar='FILENAME')
    secrets_parser.add_argument('--target-name', '-t', help='grab recipients from target name')
    secrets_parser.add_argument('--inventory-path', default='./inventory',
                                help='set inventory path, default is "./inventory"')
    secrets_parser.add_argument('--recipients', '-R', help='set recipients',
                                type=str, nargs='+', default=[], metavar='RECIPIENT')
    secrets_parser.add_argument('--secrets-path', help='set secrets path, default is "./secrets"',
                                default='./secrets',)
    secrets_parser.add_argument('--backend', help='set secrets backend, default is "gpg"',
                                type=str, choices=('gpg',), default='gpg')
    secrets_parser.add_argument('--verbose', '-v',
                                help='set verbose mode (warning: this will show sensitive data)',
                                action='store_true', default=False)
    secrets_parser.add_argument('--no-verify', help='do not verify secret hashes on reveal',
                                action='store_true', default=False)

    args = parser.parse_args()

    logger.debug('Running with args: %s', args)

    try:
        cmd = sys.argv[1]
    except IndexError:
        parser.print_help()
        sys.exit(1)

    if cmd == 'eval':
        file_path = args.jsonnet_file
        search_path = os.path.abspath(args.search_path)
        ext_vars = {}
        if args.vars:
            ext_vars = dict(var.split('=') for var in args.vars)
        json_output = None
        _search_imports = lambda cwd, imp: search_imports(cwd, imp, search_path)
        json_output = jsonnet_file(file_path, import_callback=_search_imports,
                                   native_callbacks=resource_callbacks(search_path),
                                   ext_vars=ext_vars)
        if args.output == 'yaml':
            json_obj = json.loads(json_output)
            yaml.safe_dump(json_obj, sys.stdout, default_flow_style=False)
        elif json_output:
            print(json_output)

    elif cmd == 'compile':
        if args.verbose:
            logging.basicConfig(level=logging.DEBUG,
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        elif args.quiet:
            logging.basicConfig(level=logging.CRITICAL, format="%(message)s")
        else:
            logging.basicConfig(level=logging.INFO, format="%(message)s")
        search_path = os.path.abspath(args.search_path)
        gpg_obj = secret_gpg_backend()

        if not args.ignore_version_check:
            check_version()

        compile_targets(args.inventory_path, search_path, args.output_path,
                        args.parallelism, args.targets,
                        prune=(not args.no_prune), secrets_path=args.secrets_path,
                        secrets_reveal=args.reveal, gpg_obj=gpg_obj, indent=args.indent)

        if not args.ignore_version_check:
            save_version()

    elif cmd == 'inventory':
        if args.pattern and args.target_name == '':
            parser.error("--pattern requires --target_name")
        try:
            logging.basicConfig(level=logging.INFO, format="%(message)s")
            inv = inventory_reclass(args.inventory_path)
            if args.target_name != '':
                inv = inv['nodes'][args.target_name]
                if args.pattern != '':
                    pattern = args.pattern.split(".")
                    inv = deep_get(inv, pattern)
            if args.flat:
                inv = flatten_dict(inv)
                yaml.dump(inv, sys.stdout, width=10000)
            else:
                yaml.dump(inv, sys.stdout, Dumper=PrettyDumper, default_flow_style=False)
        except Exception as e:
            if not isinstance(e, KapitanError):
                logger.error("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                traceback.print_exc()
            sys.exit(1)

    elif cmd == 'searchvar':
        searchvar(args.searchvar, args.inventory_path)

    elif cmd == 'secrets':
        if args.verbose:
            logging.basicConfig(level=logging.DEBUG,
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        else:
            logging.basicConfig(level=logging.INFO, format="%(message)s")
        gpg_obj = secret_gpg_backend()
        if args.write is not None:
            if args.file is None:
                parser.error('--file is required with --write')
            data = None
            recipients = [dict((("name", name),)) for name in args.recipients]
            if args.target_name:
                inv = inventory_reclass(args.inventory_path)
                recipients = inv['nodes'][args.target_name]['parameters']['kapitan']['secrets']['recipients']
            if args.file == '-':
                data = ''
                for line in sys.stdin:
                    data += line
            else:
                with open(args.file) as fp:
                    data = fp.read()
            secret_gpg_write(gpg_obj, args.secrets_path, args.write, data, args.base64, recipients)
        elif args.reveal:
            if args.file is None:
                parser.error('--file is required with --reveal')
            if args.file == '-':
                secret_gpg_reveal_raw(gpg_obj, args.secrets_path, None, verify=(not args.no_verify))
            elif args.file:
                if os.path.isfile(args.file):
                    out = secret_gpg_reveal_file(gpg_obj, args.secrets_path, args.file,
                                                 verify=(not args.no_verify))
                    sys.stdout.write(out)
                elif os.path.isdir(args.file):
                    secret_gpg_reveal_dir(gpg_obj, args.secrets_path, args.file,
                                          verify=(not args.no_verify))
        elif args.update:
            # update recipients for secret tag
            recipients = [dict([("name", name), ]) for name in args.recipients]
            if args.target_name:
                inv = inventory_reclass(args.inventory_path)
                recipients = inv['nodes'][args.target_name]['parameters']['kapitan']['secrets']['recipients']
            secret_gpg_update_recipients(gpg_obj, args.secrets_path, args.update, recipients)
        elif args.update_targets or args.validate_targets:
            # update recipients for all secrets in secrets_path
            # use --secrets-path to set scanning path
            inv = inventory_reclass(args.inventory_path)
            targets = set(inv['nodes'].keys())
            secrets_path = os.path.abspath(args.secrets_path)
            target_token_paths = search_target_token_paths(secrets_path, targets)
            ret_code = 0
            for target_name, token_paths in target_token_paths.items():
                try:
                    recipients = inv['nodes'][target_name]['parameters']['kapitan']['secrets']['recipients']
                    for token_path in token_paths:
                        target_fingerprints = set(lookup_fingerprints(gpg_obj, recipients))
                        secret_fingerprints = set(secret_gpg_raw_read_fingerprints(secrets_path, token_path))
                        if target_fingerprints != secret_fingerprints:
                            if args.validate_targets:
                                logger.info("%s recipient mismatch", token_path)
                                ret_code = 1
                            else:
                                new_recipients = [dict([("fingerprint", f), ]) for f in target_fingerprints]
                                secret_gpg_update_recipients(gpg_obj, secrets_path, token_path, new_recipients)
                except KeyError:
                    logger.debug("secret_gpg_update_target: target: %s has no inventory recipients, skipping",
                                 target_name)
            sys.exit(ret_code)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

import pyparsing as pp

STR = 1
REF = 2
INV = 3

def _string(string, location, tokens):
    token = tokens[0]
    tokens[0] = (STR, token)

def _reference(string, location, tokens):
    token = list(tokens[0])
    tokens[0] = (REF, token)

def _invquery(string, location, tokens):
    token = list(tokens[0])
    tokens[0] = (INV, token)

def get_ref_parser(escape_character, reference_sentinels, export_sentinels):
    _ESCAPE = escape_character
    _DOUBLE_ESCAPE = _ESCAPE + _ESCAPE

    _REF_OPEN = reference_sentinels[0]
    _REF_CLOSE = reference_sentinels[1]
    _REF_CLOSE_FIRST = _REF_CLOSE[0]
    _REF_ESCAPE_OPEN = _ESCAPE + _REF_OPEN
    _REF_ESCAPE_CLOSE = _ESCAPE + _REF_CLOSE
    _REF_DOUBLE_ESCAPE_OPEN = _DOUBLE_ESCAPE + _REF_OPEN
    _REF_DOUBLE_ESCAPE_CLOSE = _DOUBLE_ESCAPE + _REF_CLOSE
    _REF_EXCLUDES = _ESCAPE + _REF_OPEN + _REF_CLOSE

    _INV_OPEN = export_sentinels[0]
    _INV_CLOSE = export_sentinels[1]
    _INV_CLOSE_FIRST = _INV_CLOSE[0]
    _INV_ESCAPE_OPEN = _ESCAPE + _INV_OPEN
    _INV_ESCAPE_CLOSE = _ESCAPE + _INV_CLOSE
    _INV_DOUBLE_ESCAPE_OPEN = _DOUBLE_ESCAPE + _INV_OPEN
    _INV_DOUBLE_ESCAPE_CLOSE = _DOUBLE_ESCAPE + _INV_CLOSE
    _INV_EXCLUDES = _ESCAPE + _INV_OPEN + _INV_CLOSE

    _EXCLUDES = _ESCAPE + _REF_OPEN + _REF_CLOSE + _INV_OPEN + _INV_CLOSE

    double_escape = pp.Combine(pp.Literal(_DOUBLE_ESCAPE) + pp.MatchFirst([pp.FollowedBy(_REF_OPEN), pp.FollowedBy(_REF_CLOSE),
                               pp.FollowedBy(_INV_OPEN), pp.FollowedBy(_INV_CLOSE)])).setParseAction(pp.replaceWith(_ESCAPE))

    ref_open = pp.Literal(_REF_OPEN).suppress()
    ref_close = pp.Literal(_REF_CLOSE).suppress()
    ref_not_open = ~pp.Literal(_REF_OPEN) + ~pp.Literal(_REF_ESCAPE_OPEN) + ~pp.Literal(_REF_DOUBLE_ESCAPE_OPEN)
    ref_not_close = ~pp.Literal(_REF_CLOSE) + ~pp.Literal(_REF_ESCAPE_CLOSE) + ~pp.Literal(_REF_DOUBLE_ESCAPE_CLOSE)
    ref_escape_open = pp.Literal(_REF_ESCAPE_OPEN).setParseAction(pp.replaceWith(_REF_OPEN))
    ref_escape_close = pp.Literal(_REF_ESCAPE_CLOSE).setParseAction(pp.replaceWith(_REF_CLOSE))
    ref_text = pp.CharsNotIn(_REF_EXCLUDES) | pp.CharsNotIn(_REF_CLOSE_FIRST, exact=1)
    ref_content = pp.Combine(pp.OneOrMore(ref_not_open + ref_not_close + ref_text))
    ref_string = pp.MatchFirst([double_escape, ref_escape_open, ref_escape_close, ref_content]).setParseAction(_string)
    ref_item = pp.Forward()
    ref_items = pp.OneOrMore(ref_item)
    reference = (ref_open + pp.Group(ref_items) + ref_close).setParseAction(_reference)
    ref_item << (reference | ref_string)

    inv_open = pp.Literal(_INV_OPEN).suppress()
    inv_close = pp.Literal(_INV_CLOSE).suppress()
    inv_not_open = ~pp.Literal(_INV_OPEN) + ~pp.Literal(_INV_ESCAPE_OPEN) + ~pp.Literal(_INV_DOUBLE_ESCAPE_OPEN)
    inv_not_close = ~pp.Literal(_INV_CLOSE) + ~pp.Literal(_INV_ESCAPE_CLOSE) + ~pp.Literal(_INV_DOUBLE_ESCAPE_CLOSE)
    inv_escape_open = pp.Literal(_INV_ESCAPE_OPEN).setParseAction(pp.replaceWith(_INV_OPEN))
    inv_escape_close = pp.Literal(_INV_ESCAPE_CLOSE).setParseAction(pp.replaceWith(_INV_CLOSE))
    inv_text = pp.CharsNotIn(_INV_CLOSE_FIRST)
    inv_content = pp.Combine(pp.OneOrMore(inv_not_close + inv_text))
    inv_string = pp.MatchFirst([double_escape, inv_escape_open, inv_escape_close, inv_content]).setParseAction(_string)
    inv_items = pp.OneOrMore(inv_string)
    export = (inv_open + pp.Group(inv_items) + inv_close).setParseAction(_invquery)

    text = pp.CharsNotIn(_EXCLUDES) | pp.CharsNotIn('', exact=1)
    content = pp.Combine(pp.OneOrMore(ref_not_open + inv_not_open + text))
    string = pp.MatchFirst([double_escape, ref_escape_open, inv_escape_open, content]).setParseAction(_string)

    item = reference | export | string
    line = pp.OneOrMore(item) + pp.StringEnd()
    return line

def get_simple_ref_parser(escape_character, reference_sentinels, export_sentinels):
    _ESCAPE = escape_character
    _REF_OPEN = reference_sentinels[0]
    _REF_CLOSE = reference_sentinels[1]
    _INV_OPEN = export_sentinels[0]
    _INV_CLOSE = export_sentinels[1]
    _EXCLUDES = _ESCAPE + _REF_OPEN + _REF_CLOSE + _INV_OPEN + _INV_CLOSE

    string = pp.CharsNotIn(_EXCLUDES).setParseAction(_string)
    ref_open = pp.Literal(_REF_OPEN).suppress()
    ref_close = pp.Literal(_REF_CLOSE).suppress()
    reference = (ref_open + pp.Group(string) + ref_close).setParseAction(_reference)
    line = pp.StringStart() + pp.Optional(string) + reference + pp.Optional(string) + pp.StringEnd()
    return line



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from .item import Item
from reclass.defaults import REFERENCE_SENTINELS
from reclass.settings import Settings
from reclass.utils.dictpath import DictPath
from reclass.errors import ResolveError


class RefItem(Item):

    def __init__(self, items, settings):
        self.type = Item.REFERENCE
        self._settings = settings
        self._items = items
        self._refs = []
        self._allRefs = False
        self.assembleRefs()

    def assembleRefs(self, context={}):
        self._refs = []
        self._allRefs = True
        for item in self._items:
            if item.has_references():
                item.assembleRefs(context)
                self._refs.extend(item.get_references())
                if item.allRefs() == False:
                    self._allRefs = False
        try:
            strings = [ str(i.render(context, None)) for i in self._items ]
            value = "".join(strings)
            self._refs.append(value)
        except ResolveError as e:
            self._allRefs = False

    def contents(self):
        return self._items

    def allRefs(self):
        return self._allRefs

    def has_references(self):
        return len(self._refs) > 0

    def get_references(self):
        return self._refs

    def _resolve(self, ref, context):
        path = DictPath(self._settings.delimiter, ref)
        try:
            return path.get_value(context)
        except (KeyError, TypeError) as e:
            raise ResolveError(ref)

    def render(self, context, inventory):
        if len(self._items) == 1:
            return self._resolve(self._items[0].render(context, inventory), context)
        strings = [ str(i.render(context, inventory)) for i in self._items ]
        return self._resolve("".join(strings), context)

    def __repr__(self):
        return 'RefItem(%r)' % self._items

    def __str__(self):
        strings = [ str(i) for i in self._items ]
        return '{0}{1}{2}'.format(REFERENCE_SENTINELS[0], ''.join(strings), REFERENCE_SENTINELS[1])



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from .item import Item
from reclass.settings import Settings

class ListItem(Item):

    def __init__(self, item, settings):
        self.type = Item.LIST
        self._list = item
        self._settings = settings

    def contents(self):
        return self._list

    def is_container(self):
        return True

    def render(self, context, inventory):
        return self._list

    def merge_over(self, item):
        if item.type == Item.LIST:
            item._list.extend(self._list)
            return item
        elif item.type == Item.SCALAR:
            if item.contents() is None:
                return self
            elif self._settings.allow_list_over_scalar:
                self._list.insert(0, item.contents())
                return self
            else:
                raise TypeError('allow list over scalar = False: cannot merge %s onto %s' % (repr(self), repr(item)))
        raise TypeError('Cannot merge %s over %s' % (repr(self), repr(item)))

    def __repr__(self):
        return 'ListItem(%r)' % (self._list)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from __future__ import print_function

import copy
import sys

from reclass.errors import ResolveError

class ValueList(object):

    def __init__(self, value, settings):
        self._settings = settings
        self._refs = []
        self._allRefs = True
        self._values = [ value ]
        self._inv_refs = []
        self._has_inv_query = False
        self._ignore_failed_render = False
        self._update()

    def append(self, value):
        self._values.append(value)
        self._update()

    def extend(self, values):
        self._values.extend(values._values)
        self._update()

    def _update(self):
        self.assembleRefs()
        self._check_for_inv_query()

    def has_references(self):
        return len(self._refs) > 0

    def has_inv_query(self):
        return self._has_inv_query

    def get_inv_references(self):
        return self._inv_refs

    def is_complex(self):
        return (self.has_references() | self.has_inv_query())

    def get_references(self):
        return self._refs

    def allRefs(self):
        return self._allRefs

    def ignore_failed_render(self):
        return self._ignore_failed_render

    def _check_for_inv_query(self):
        self._has_inv_query = False
        self._ignore_failed_render = True
        for value in self._values:
            if value.has_inv_query():
                self._inv_refs.extend(value.get_inv_references)
                self._has_inv_query = True
                if vale.ignore_failed_render() is False:
                    self._ignore_failed_render = False
        if self._has_inv_query is False:
            self._ignore_failed_render = False

    def assembleRefs(self, context={}):
        self._refs = []
        self._allRefs = True
        for value in self._values:
            value.assembleRefs(context)
            if value.has_references():
                self._refs.extend(value.get_references())
            if value.allRefs() is False:
                self._allRefs = False

    def merge(self):
        output = None
        for n, value in enumerate(self._values):
            if output is None:
                output = value
            else:
                output = value.merge_over(output)
        return output

    def render(self, context, inventory):
        from reclass.datatypes.parameters import Parameters

        output = None
        deepCopied = False
        last_error = None
        for n, value in enumerate(self._values):
            try:
                new = value.render(context, inventory)
            except ResolveError as e:
                if self._settings.ignore_overwritten_missing_references and not isinstance(output, (dict, list)) and n != (len(self._values)-1):
                    new = None
                    last_error = e
                    print("[WARNING] Reference '%s' undefined" % str(value), file=sys.stderr)
                else:
                    raise e

            if output is None or value.overwrite:
                output = new
                deepCopied = False
            else:
                if isinstance(output, dict) and isinstance(new, dict):
                    p1 = Parameters(output, self._settings, None, parse_strings=False)
                    p2 = Parameters(new, self._settings, None, parse_strings=False)
                    p1.merge(p2)
                    output = p1.as_dict()
                    continue
                elif isinstance(output, list) and isinstance(new, list):
                    if not deepCopied:
                        output = copy.deepcopy(output)
                        deepCopied = True
                    output.extend(new)
                    continue
                elif isinstance(output, (dict, list)) or isinstance(new, (dict, list)):
                    raise TypeError('Cannot merge %s over %s' % (repr(self._values[n]), repr(self._values[n-1])))
                else:
                    output = new
                    deepCopied = False

        if isinstance(output, (dict, list)) and last_error is not None:
            raise last_error

        return output

    def __repr__(self):
        return 'ValueList(%r)' % self._values



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from reclass.utils.dictpath import DictPath

class Item(object):

    COMPOSITE = 1
    DICTIONARY = 2
    INV_QUERY = 3
    LIST = 4
    REFERENCE = 5
    SCALAR = 6

    def allRefs(self):
        return True

    def has_references(self):
        return False

    def has_inv_query(self):
        return False

    def is_container(self):
        return False

    def is_complex(self):
        return (self.has_references() | self.has_inv_query())

    def contents(self):
        msg = "Item class {0} does not implement contents()"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def merge_over(self, item):
        msg = "Item class {0} does not implement merge_over()"
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def render(self, context, exports):
        msg = "Item class {0} does not implement render()"
        raise NotImplementedError(msg.format(self.__class__.__name__))



#!/usr/bin/env python3.6
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"jsonnet tests"

import unittest
from kapitan.resources import yaml_dump


class JsonnetNativeFuncsTest(unittest.TestCase):
    def test_yaml_dump(self):
        "dump json string to yaml"
        yaml = yaml_dump("{\"key\":\"value\"}")
        self.assertEqual(yaml, "key: value\n")



#!/usr/bin/env python3.6
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"kapitan targets"

from six import string_types

import logging
import os
import errno
import hashlib
import ujson as json
import re
import shutil
import sys
from functools import partial
import multiprocessing
import traceback
import tempfile
import jsonschema
import yaml
import time

from kapitan.resources import search_imports, resource_callbacks, inventory, inventory_reclass
from kapitan.utils import jsonnet_file, prune_empty, render_jinja2, PrettyDumper
from kapitan.secrets import secret_gpg_raw_read, secret_token_from_tag, secret_token_attributes
from kapitan.secrets import SECRET_TOKEN_TAG_PATTERN, secret_gpg_read, secret_gpg_write
from kapitan.secrets import secret_gpg_exists, secret_gpg_write_function
from kapitan.errors import KapitanError, CompileError, SecretError
from kapitan import cached

logger = logging.getLogger(__name__)


def compile_targets(inventory_path, search_path, output_path, parallel, targets, **kwargs):
    """
    Searches and loads target files, and runs compile_target_file() on a
    multiprocessing pool with parallel number of processes.
    kwargs are passed to compile_target()
    """
    # temp_path will hold compiled items
    temp_path = tempfile.mkdtemp(suffix='.kapitan')

    target_objs = load_target_inventory(inventory_path, targets)

    pool = multiprocessing.Pool(parallel)
    # append "compiled" to output_path so we can safely overwrite it
    compile_path = os.path.join(output_path, "compiled")
    worker = partial(compile_target, search_path=search_path, compile_path=temp_path, **kwargs)

    try:
        if target_objs == []:
            logger.error("Error: no targets found")
            raise KapitanError("Error: no targets found")
        # compile_target() returns None on success
        # so p is only not None when raising an exception
        [p.get() for p in pool.imap_unordered(worker, target_objs) if p]

        if not os.path.exists(compile_path):
            os.makedirs(compile_path)

        # if '-t' is set on compile, only override selected targets
        if targets:
            for target in targets:
                compile_path_target = os.path.join(compile_path, target)
                temp_path_target = os.path.join(temp_path, target)

                if not os.path.exists(compile_path_target):
                    os.makedirs(compile_path_target)

                shutil.rmtree(compile_path_target)
                shutil.copytree(temp_path_target, compile_path_target)
                logger.debug("Copied %s into %s", temp_path_target, compile_path_target)
        # otherwise override all targets
        else:
            shutil.rmtree(compile_path)
            shutil.copytree(temp_path, compile_path)
            logger.debug("Copied %s into %s", temp_path, compile_path)

    except Exception as e:
        # if compile worker fails, terminate immediately
        pool.terminate()
        pool.join()
        logger.debug("Compile pool terminated")
        # only print traceback for errors we don't know about
        if not isinstance(e, KapitanError):
            logger.error("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            traceback.print_exc()
        sys.exit(1)
    finally:
        shutil.rmtree(temp_path)
        logger.debug("Removed %s", temp_path)


def load_target_inventory(inventory_path, targets):
    "retuns a list of target objects from the inventory"
    target_objs = []
    inv = inventory_reclass(inventory_path)

    targets_list = inv['nodes']
    # if '-t' is set on compile, only loop through selected targets
    if targets:
        targets_list = targets

    for target_name in targets_list:
        try:
            target_obj = inv['nodes'][target_name]['parameters']['kapitan']
            valid_target_obj(target_obj)
            logger.debug("load_target_inventory: found valid kapitan target %s", target_name)
            target_objs.append(target_obj)
        except KeyError:
            logger.debug("load_target_inventory: target %s has no kapitan compile obj", target_name)
            pass

    return target_objs


def compile_target(target_obj, search_path, compile_path, **kwargs):
    """
    Compiles target_obj and writes to compile_path
    """
    start = time.time()

    ext_vars = target_obj["vars"]
    target_name = ext_vars["target"]
    compile_obj = target_obj["compile"]

    for obj in compile_obj:
        input_type = obj["input_type"]
        input_paths = obj["input_paths"]
        output_path = obj["output_path"]
        if input_type == "jsonnet":
            _compile_path = os.path.join(compile_path, target_name, output_path)
            # support writing to an already existent dir
            try:
                os.makedirs(_compile_path)
            except OSError as ex:
                # If directory exists, pass
                if ex.errno == errno.EEXIST:
                    pass
            output_type = obj["output_type"]  # output_type is mandatory in jsonnet
            for input_path in input_paths:
                compile_file_sp = os.path.join(search_path, input_path)
                if os.path.exists(compile_file_sp):
                    logger.debug("Compiling %s", compile_file_sp)
                    try:
                        compile_jsonnet(compile_file_sp, _compile_path, search_path,
                                        ext_vars, output=output_type, target_name=target_name,
                                        **kwargs)
                    except CompileError as e:
                        logger.error("Compile error: failed to compile target: %s",
                                     target_name)
                        raise e
                else:
                    logger.error("Compile error: input_path for target: %s not found in search_path: %s",
                                 target_name, input_path)
                    raise CompileError()

        if input_type == "jinja2":
            _compile_path = os.path.join(compile_path, target_name, output_path)
            # support writing to an already existent dir
            try:
                os.makedirs(_compile_path)
            except OSError as ex:
                # If directory exists, pass
                if ex.errno == errno.EEXIST:
                    pass
            for input_path in input_paths:
                compile_path_sp = os.path.join(search_path, input_path)
                if os.path.exists(compile_path_sp):
                    # copy ext_vars to dedicated jinja2 context so we can update it
                    ctx = ext_vars.copy()
                    ctx["inventory"] = inventory(search_path, target_name)
                    ctx["inventory_global"] = inventory(search_path, None)
                    try:
                        compile_jinja2(compile_path_sp, ctx, _compile_path,
                                       target_name=target_name, **kwargs)
                    except CompileError as e:
                        logger.error("Compile error: failed to compile target: %s",
                                     target_name)
                        raise e
                else:
                    logger.error("Compile error: input_path for target: %s not found in search_path: %s",
                                 target_name, input_path)
                    raise CompileError()

    logger.info("Compiled %s (%.2fs)", target_name, time.time() - start)


def compile_jinja2(path, context, compile_path, **kwargs):
    """
    Write items in path as jinja2 rendered files to compile_path.
    path can be either a file or directory.
    kwargs:
        secrets_path: default None, set to access secrets backend
        secrets_reveal: default False, set to reveal secrets on compile
        gpg_obj: default None
        target_name: default None, set to current target being compiled
    """
    secrets_path = kwargs.get('secrets_path', None)
    secrets_reveal = kwargs.get('secrets_reveal', False)
    gpg_obj = kwargs.get('gpg_obj', None)
    target_name = kwargs.get('target_name', None)

    for item_key, item_value in render_jinja2(path, context).items():
        full_item_path = os.path.join(compile_path, item_key)
        try:
            os.makedirs(os.path.dirname(full_item_path))
        except OSError as ex:
            # If directory exists, pass
            if ex.errno == errno.EEXIST:
                pass
        with CompiledFile(full_item_path, mode="w", secrets_path=secrets_path,
                          secrets_reveal=secrets_reveal, gpg_obj=gpg_obj,
                          target_name=target_name) as fp:
            fp.write(item_value["content"])
            mode = item_value["mode"]
            os.chmod(full_item_path, mode)
            logger.debug("Wrote %s with mode %.4o", full_item_path, mode)


def compile_jsonnet(file_path, compile_path, search_path, ext_vars, **kwargs):
    """
    Write file_path (jsonnet evaluated) items as files to compile_path.
    Set output to write as json or yaml
    search_path and ext_vars will be passed as parameters to jsonnet_file()
    kwargs:
        output: default 'yaml', accepts 'json'
        prune: default True, accepts False
        secrets_path: default None, set to access secrets backend
        secrets_reveal: default False, set to reveal secrets on compile
        target_name: default None, set to current target being compiled
        gpg_obj: default None
        indent: default 2
    """
    _search_imports = lambda cwd, imp: search_imports(cwd, imp, search_path)
    json_output = jsonnet_file(file_path, import_callback=_search_imports,
                               native_callbacks=resource_callbacks(search_path),
                               ext_vars=ext_vars)
    json_output = json.loads(json_output)

    output = kwargs.get('output', 'yaml')
    prune = kwargs.get('prune', True)
    secrets_path = kwargs.get('secrets_path', None)
    secrets_reveal = kwargs.get('secrets_reveal', False)
    gpg_obj = kwargs.get('gpg_obj', None)
    target_name = kwargs.get('target_name', None)
    indent = kwargs.get('indent', 2)

    if prune:
        json_output = prune_empty(json_output)
        logger.debug("Pruned output for: %s", file_path)

    for item_key, item_value in json_output.items():
        # write each item to disk
        if output == 'json':
            file_path = os.path.join(compile_path, '%s.%s' % (item_key, output))
            with CompiledFile(file_path, mode="w", secrets_path=secrets_path,
                              secrets_reveal=secrets_reveal, gpg_obj=gpg_obj,
                              target_name=target_name, indent=indent) as fp:
                fp.write_json(item_value)
                logger.debug("Wrote %s", file_path)
        elif output == 'yaml':
            file_path = os.path.join(compile_path, '%s.%s' % (item_key, "yml"))
            with CompiledFile(file_path, mode="w", secrets_path=secrets_path,
                              secrets_reveal=secrets_reveal, gpg_obj=gpg_obj,
                              target_name=target_name, indent=indent) as fp:
                fp.write_yaml(item_value)
                logger.debug("Wrote %s", file_path)
        else:
            raise ValueError('output is neither "json" or "yaml"')


def valid_target_obj(target_obj):
    """
    Validates a target_obj
    Returns a dict object if target is valid
    Otherwise raises ValidationError
    """

    schema = {
        "type": "object",
        "properties": {
            "vars": {"type": "object"},
            "secrets": {"type": "object"},
            "compile": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "input_paths": {"type": "array"},
                        "input_type": {"type": "string"},
                        "output_path": {"type": "string"},
                        "output_type": {"type": "string"},
                    },
                    "required": ["input_type", "input_paths", "output_path"],
                    "minItems": 1,
                }
            },
        },
        "required": ["compile"],
    }

    return jsonschema.validate(target_obj, schema)


class CompilingFile(object):
    def __init__(self, context, fp, **kwargs):
        self.fp = fp
        self.kwargs = kwargs

    def write(self, data):
        "write data into file"
        secrets_reveal = self.kwargs.get('secrets_reveal', False)
        if secrets_reveal:
            self.fp.write(self.sub_token_reveal_data(data))
        else:
            self.fp.write(self.sub_token_compiled_data(data))

    def write_yaml(self, obj):
        "recursively hash or reveal secrets and convert obj to yaml and write to file"
        indent = self.kwargs.get('indent', 2)
        secrets_reveal = self.kwargs.get('secrets_reveal', False)
        if secrets_reveal:
            self.sub_token_reveal_obj(obj)
        else:
            self.sub_token_compiled_obj(obj)
        yaml.dump(obj, stream=self.fp, indent=indent, Dumper=PrettyDumper, default_flow_style=False)

    def write_json(self, obj):
        "recursively hash or reveal secrets and convert obj to json and write to file"
        indent = self.kwargs.get('indent', 2)
        secrets_reveal = self.kwargs.get('secrets_reveal', False)
        if secrets_reveal:
            self.sub_token_reveal_obj(obj)
        else:
            self.sub_token_compiled_obj(obj)
        json.dump(obj, self.fp, indent=indent, escape_forward_slashes=False)

    def sub_token_compiled_obj(self, obj):
        "recursively find and replace tokens with hashed tokens in obj"
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self.sub_token_compiled_obj(v)
        elif isinstance(obj, list):
            obj = list(map(self.sub_token_compiled_obj, obj))
        elif isinstance(obj, string_types):
            obj = self.sub_token_compiled_data(obj)

        return obj

    def sub_token_reveal_obj(self, obj):
        "recursively find and reveal token tags in data"
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self.sub_token_reveal_obj(v)
        elif isinstance(obj, list):
            obj = list(map(self.sub_token_reveal_obj, obj))
        elif isinstance(obj, string_types):
            obj = self.sub_token_reveal_data(obj)

        return obj

    def hash_token_tag(self, token_tag):
        """
        suffixes a secret's hash to its tag:
        e.g:
        ?{gpg:app1/secret/1} gets replaced with
        ?{gpg:app1/secret/1:deadbeef}
        """
        secrets_path = self.kwargs.get("secrets_path", None)
        if secrets_path is None:
            raise ValueError("secrets_path not set")
        token, func = secret_token_from_tag(token_tag)
        secret_raw_obj = secret_gpg_raw_read(secrets_path, token)
        backend, token_path = secret_token_attributes(token)
        sha256 = hashlib.sha256("%s%s".encode("UTF-8") % (token_path.encode("UTF-8"),
                                                          secret_raw_obj["data"].encode("UTF-8"))).hexdigest()
        sha256 = sha256[:8]
        return "?{%s:%s:%s}" % (backend, token_path, sha256)

    def reveal_token_tag(self, token_tag):
        "reveal token_tag"
        secrets_path = self.kwargs.get("secrets_path", None)
        gpg_obj = self.kwargs.get("gpg_obj", None)
        if secrets_path is None:
            raise ValueError("secrets_path not set")
        if gpg_obj is None:
            raise ValueError("secrets_path not set")

        token, func = secret_token_from_tag(token_tag)
        return secret_gpg_read(gpg_obj, secrets_path, token)

    def sub_token_compiled_data(self, data):
        "find and replace tokens with hashed tokens in data"
        def _hash_token_tag(match_obj):
            token_tag, token, func = match_obj.groups()
            _, token_path = secret_token_attributes(token)
            secrets_path = self.kwargs.get("secrets_path", None)
            if secrets_path is None:
                raise ValueError('secrets_path not set')

            # if token secret func is defined and secret does not exist
            # write secret from func eval
            if func and not secret_gpg_exists(secrets_path, token_path):
                logger.info("Creating secret for %s:%s ...", token_path, func)
                self.target_secret_func_write(token_path, func)

            return self.hash_token_tag(token_tag)

        return re.sub(SECRET_TOKEN_TAG_PATTERN, _hash_token_tag, data)

    def sub_token_reveal_data(self, data):
        "find and reveal token tags in data"
        def _reveal_token_tag(match_obj):
            token_tag, _ = match_obj.groups()
            return self.reveal_token_tag(token_tag)

        return re.sub(SECRET_TOKEN_TAG_PATTERN, _reveal_token_tag, data)

    def target_secret_func_write(self, token_path, func):
        "write a target secret for token with data from func"
        target_name = self.kwargs.get('target_name', None)
        gpg_obj = self.kwargs.get("gpg_obj", None)
        secrets_path = self.kwargs.get("secrets_path", None)
        target_inv = cached.inv['nodes'].get(target_name, None)
        if target_name is None:
            raise ValueError('target_name not set')
        if gpg_obj is None:
            raise ValueError('gpg_obj not set')
        if target_inv is None:
            raise ValueError('target_inv not set')
        if secrets_path is None:
            raise ValueError('secrets_path not set')
        recipients = target_inv['parameters']['kapitan']['secrets']['recipients']

        secret_gpg_write_function(gpg_obj, secrets_path, token_path, func, recipients)


class CompiledFile(object):
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs
        self.fp = None

    def __enter__(self):
        mode = self.kwargs.get("mode", "r")
        self.fp = open(self.name, mode)
        return CompilingFile(self, self.fp, **self.kwargs)

    def __exit__(self, *args):
        self.fp.close()
